'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var findCacheDir = _interopDefault(require('find-cache-dir'));
var path = _interopDefault(require('path'));
var mkdirp = _interopDefault(require('mkdirp'));
var t = require('@babel/types');
var core = require('@babel/core');
var register = require('@babel/register');
var register__default = _interopDefault(register);
var evaluatePath = _interopDefault(require('babel-helper-evaluate-path'));
var resolve = _interopDefault(require('resolve'));
var taddy = require('taddy');
var core$1 = require('@taddy/core');
var helperModuleImports = require('@babel/helper-module-imports');
var fs = _interopDefault(require('fs'));
var stringHash = _interopDefault(require('string-hash'));

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function getType(symbol) {
  return symbol === null || symbol === void 0 ? void 0 : symbol.getTypeAtLocation(symbol.getDeclarations()[0]);
}
class DynamicType {
  constructor({
    value,
    type
  }) {
    this.value = value;
    this.type = type;
  }

}
class UnionType extends Array {}

function isRecord(type) {
  var _type$getAliasSymbol;

  return (type === null || type === void 0 ? void 0 : (_type$getAliasSymbol = type.getAliasSymbol()) === null || _type$getAliasSymbol === void 0 ? void 0 : _type$getAliasSymbol.getEscapedName()) === 'Record';
}

const STATIC_KEY = /*#__PURE__*/Symbol('__TADDY_STATIC_KEY__');
function isStaticValue(value) {
  if (!value || typeof value !== 'object') {
    return true;
  }

  if (value instanceof UnionType) {
    return false;
  }

  return value[STATIC_KEY] || false;
}

function filterValue(value) {
  // filter vendor prefixes
  return String(value).startsWith('-');
}

function parseValue(type) {
  var _value;

  if (!type) return null;

  if (type.isAny() || type.isUnknown()) {
    return new DynamicType({
      value: null,
      type
    });
  }

  if (isRecord(type)) {
    return parseRecord(type);
  }

  if (type.isArray()) {
    return parseArray(type);
  }

  if (type.isUnion()) {
    return new UnionType(...type.getUnionTypes().map(parseValue));
  }

  if (type.isObject()) {
    return parseObject(type);
  }

  let value = type === null || type === void 0 ? void 0 : type.getText();
  /**
   * Firstly, exclude types like "string & {}"
   *
   * TODO: think about other intersection types
   */

  if (!value || ((_value = value) === null || _value === void 0 ? void 0 : _value.includes('&'))) {
    return null;
  }

  try {
    value = JSON.parse(value);
  } catch (e) {
    return new DynamicType({
      value,
      type
    });
  }

  if (filterValue(value)) return null;
  return value;
}

function parseMixinParameters(params) {
  const type = getType(params);
  return parseValue(type);
}

function parseMixin(mixin) {
  const type = getType(mixin);

  if ((type === null || type === void 0 ? void 0 : type.getCallSignatures().length) === 0) {
    return parseMixinParameters(mixin);
  }

  return type === null || type === void 0 ? void 0 : type.getCallSignatures()[0].getParameters().map(parseMixinParameters);
}

function parseRecord(type) {
  const [k, v] = type.getAliasTypeArguments();
  const key = parseValue(k);
  const value = parseValue(v);

  if (!Array.isArray(key)) {
    return {
      [key]: value,
      [STATIC_KEY]: value[STATIC_KEY]
    };
  }

  return key.map(prop => ({
    [prop]: value,
    [STATIC_KEY]: value[STATIC_KEY]
  }));
} // TODO: improve the array parsing (such as not just union types)


function parseArray(type) {
  const elements = [];
  elements[STATIC_KEY] = true;

  function applyElement(element) {
    const value = parseValue(element);
    if (!value) return;
    const isStatic = isStaticValue(value);
    elements[STATIC_KEY] = elements[STATIC_KEY] && isStatic;
    elements.push(isStatic ? value : new Set([].concat(value).filter(Boolean)));
  }

  const elementType = type.getArrayElementTypeOrThrow();

  for (let element of elementType.getUnionTypes() || [elementType]) {
    applyElement(element);
  }

  return elements;
}

function parseObject(type) {
  const props = {};
  let isStaticObject = true;

  for (const propSymbol of (_type$getProperties = type === null || type === void 0 ? void 0 : type.getProperties()) !== null && _type$getProperties !== void 0 ? _type$getProperties : []) {
    var _type$getProperties;

    const name = propSymbol.getEscapedName();
    const value = parseMixin(propSymbol);
    if (!value) continue;
    const isStatic = isStaticValue(value);
    isStaticObject = isStaticObject && isStatic;
    props[name] = isStatic ? value : new Set([].concat(value).filter(Boolean));
  }

  props[STATIC_KEY] = isStaticObject;
  return props;
}
class TSProcessor {
  constructor(config) {
    const {
      Project
    } = require('ts-morph');

    this.project = new Project(_extends({
      tsConfigFilePath: path.join(process.cwd(), 'tsconfig.json')
    }, config));
  }

  getTypeAtPos(filename, code, start, end) {
    let sourceFile = this.project.getSourceFile(filename);

    if (sourceFile && sourceFile.getFullText() !== code) {
      sourceFile.replaceText([0, sourceFile.getEnd()], code);
    } else {
      sourceFile = this.project.createSourceFile(filename, code, {
        overwrite: true
      });
    }

    const typeNode = sourceFile.getDescendantAtStartWithWidth(start, end - start);
    return typeNode;
  }

}

const MACRO_NAME = 'taddy.macro';
const PACKAGE_NAME = 'taddy';
// export const config = lilconfigSync(PACKAGE_NAME, {
//     searchPlaces: [`${PACKAGE_NAME}.config.ts`],
//     loaders: {
//         '.ts': loadConfig,
//     },
// }).search()?.config;

const cacheDir = /*#__PURE__*/findCacheDir({
  name: PACKAGE_NAME
}) || /*#__PURE__*/path.join( /*#__PURE__*/require.resolve('taddy'), '.cache');
const getCachedModuleFilepath = filename => {
  return `.cache/${PACKAGE_NAME}/${filename}`;
};

try {
  mkdirp.sync(cacheDir);
} catch (error) {
  // TODO: handle this error
  console.error('TADDY', 'can not create cache dir', error);
}

const cache = /*#__PURE__*/new WeakMap();
class BindingError extends Error {}
function findBindings(currentPath) {
  function visitIdentifier(referentPath, path, bindingSet) {
    if (path.parentPath.isMemberExpression()) {
      // @ts-ignore
      path = path.parentPath.get('object');
    }

    const {
      parentPath,
      node
    } = path;

    if (!(parentPath.isExpression() || parentPath.isSpreadElement() || parentPath.isObjectProperty() && (parentPath.node.value === node || parentPath.node.computed))) {
      return;
    }

    if (!('name' in node && typeof node.name === 'string')) return;
    const {
      name
    } = node;
    const sharedBinding = currentPath.scope.getBinding(name);
    const binding = referentPath.scope.getBinding(name);
    if (!binding) return;
    if (sharedBinding !== binding) return;

    if (binding.constantViolations.length > 0) {
      throw new BindingError('CONSTANT VIOLATION');
    }

    const {
      path: bindingPath
    } = binding;

    if (bindingSet.has(binding)) {
      return;
    }

    if (bindingPath.parentPath.isFunctionExpression() && bindingPath.listKey === 'params') {
      throw new BindingError('FUNCTION ARGUMENT');
    }

    bindingSet.add(binding);

    if (bindingPath.isImportSpecifier()) {
      return;
    }

    (traverse(bindingPath) || []).forEach(x => bindingSet.add(x));
  }

  const visited = new Set();

  function traverse(referentPath) {
    if (referentPath.isObjectPattern()) {
      throw new BindingError('OBJECT PATTERN');
    }

    if (cache.has(referentPath)) {
      return cache.get(referentPath);
    } // we should not revisit the paths, otherwise something went wrong


    if (visited.has(referentPath)) {
      throw new BindingError('INFINITE LOOP');
    }

    visited.add(referentPath);
    const bindingSet = new Set();

    if (referentPath.isIdentifier()) {
      visitIdentifier(referentPath, referentPath, bindingSet);
    }

    referentPath.traverse({
      Identifier(path) {
        visitIdentifier(referentPath, path, bindingSet);
      }

    });
    cache.set(referentPath, bindingSet);
    return bindingSet;
  }

  return traverse(currentPath);
}

function getPathsByBindings(bindings) {
  const paths = new Set();

  for (let x of bindings) {
    if (x.path.isImportSpecifier() || x.path.isVariableDeclarator()) {
      paths.add(x.path.parentPath);
      continue;
    }

    paths.add(x.path);
  }

  return paths;
}

function buildCodeByPath(path) {
  return Array.from(getPathsByBindings(findBindings(path))).sort((a, b) => a.node.start - b.node.start).map(x => x.toString()).join('\n');
}

const DEFAULT_PRESETS = ['@babel/preset-typescript', '@babel/preset-react', ['@babel/preset-env', {
  targets: {
    node: 'current'
  }
}]];
const EXTENSIONS = ['.es6', '.es', '.tsx', '.ts', '.jsx', '.js', '.mjs'];
const macroRe = /*#__PURE__*/new RegExp( /*#__PURE__*/MACRO_NAME.replace('.', '\\.'), 'g');
const EVAL_FILENAME_POSTFIX = '@__TADDY_EVALUATE__';
function isTaddyEvaluation(state) {
  return state.filename.includes(EVAL_FILENAME_POSTFIX);
}
function evaluate(currentPath) {
  const result = evaluatePath(currentPath);

  if (result.confident) {
    return {
      value: result.value
    };
  }

  try {
    const callbackName = '__taddy__';
    const content = buildCodeByPath(currentPath) // hack to avoid processing by babel-macro
    .replace(macroRe, PACKAGE_NAME).concat(`${callbackName}(${currentPath.toString()})`);
    const {
      opts
    } = currentPath.hub.file;
    const ext = path.extname(opts.filename);
    const basename = path.basename(opts.filename, ext);
    const dirname = path.dirname(opts.filename);
    const options = {
      babelrc: false,
      configFile: false,
      filename: path.join(dirname, basename + EVAL_FILENAME_POSTFIX + ext),
      plugins: [
        /*...opts.plugins*/
      ],
      presets: [
      /*...opts.presets*/
      ...DEFAULT_PRESETS]
    }; // console.log({content});

    const {
      code
    } = core.transformSync(content, options) || {};
    if (!code) return {};
    const exec = new Function('require', callbackName, code);
    register__default(_extends({}, options, {
      extensions: EXTENSIONS
    }));
    let value;
    exec(filepath => {
      if (filepath === PACKAGE_NAME) return taddy;
      return require(resolve.sync(filepath, {
        extensions: EXTENSIONS,
        basedir: path.dirname(opts.filename)
      }));
    }, result => {
      value = result;
    });
    register.revert();
    return {
      value
    };
  } catch (error) {
    // console.log({error});
    return {
      error
    };
  }
}

function getObjectPropertyKey(path) {
  if (t.isIdentifier(path.node.key)) {
    return path.node.key.name;
  }

  if (t.isLiteral(path.node.key) && 'value' in path.node.key) {
    return path.node.key.value;
  }

  return null;
}

function mergeObjectProperties(properties) {
  const map = new Map();

  function move(key, value) {
    map.delete(key);
    map.set(key, value);
  }

  for (const x of properties) {
    if (!t.isObjectProperty(x)) {
      move(x, x);
      continue;
    }

    const {
      key
    } = x;

    if (t.isIdentifier(key)) {
      move(key.name, x);
      continue;
    }

    if (t.isStringLiteral(key)) {
      move(key.value, x);
      continue;
    }

    move(x, x);
  } // TODO: there was a problem with typescript compiler with the variant [...map.values()]


  return Array.from(map.values());
}

function isTaddy(binding) {
  const {
    path
  } = binding;
  const {
    parentPath
  } = path;
  return parentPath.isImportDeclaration() && parentPath.node.source.value === 'taddy';
}

function optimizeBindings(referentPath) {
  let bindings;

  try {
    bindings = findBindings(referentPath);
  } catch (error) {
    console.log('ERROR', {
      error
    });
    return;
  }

  for (let binding of bindings) {
    if (isTaddy(binding)) continue;
    binding.references--;
    optimizeBindings(binding.path);

    if (binding.references === 0) {
      binding.path.remove();
      const {
        parentPath
      } = binding.path;

      if (parentPath.isImportDeclaration() && parentPath.node.specifiers.length === 0) {
        parentPath.remove();
        continue;
      }
    }
  }
}

function optimizeStaticStyles(path) {
  const quasis = [];
  const expressions = [];
  const props = [];

  for (const propPath of path.get('properties')) {
    if (!propPath.isObjectProperty()) {
      throw new Error('Cant optimize this path');
    }

    const key = String(getObjectPropertyKey(propPath));
    const valuePath = propPath.get('value');

    if (key === 'className') {
      expressions.push(valuePath.node);
      continue;
    }

    if (key === 'style' || key === core$1.VARS_KEY) {
      props.push(propPath.node);
      continue;
    }

    if (valuePath.isStringLiteral()) {
      quasis.push(key + valuePath.node.value);
      continue;
    }

    if (valuePath.isBooleanLiteral()) {
      quasis.push(key);
      continue;
    }

    expressions.push(t.binaryExpression('+', t.stringLiteral(key), valuePath.node));
  }

  let classes = t.stringLiteral(quasis.join(' '));

  if (expressions.length !== 0) {
    classes.value += classes.value ? ' ' : '';
    const left = classes;
    const right = expressions.length === 1 ? expressions[0] : t.templateLiteral([t.templateElement({
      raw: ''
    })].concat([...Array(expressions.length - 1)].fill(t.templateElement({
      raw: ' '
    }))).concat(t.templateElement({
      raw: ''
    })), expressions);
    classes = t.binaryExpression('+', left, right);
  }

  if (props.length === 0) {
    path.replaceWith(classes);
    return;
  }

  path.node.properties = props.concat(t.objectProperty(t.identifier('className'), classes));
}

function getLiteralValue(path) {
  // eslint-disable-next-line no-eval
  return eval(path.toString());
}

function getHashedName(key, {
  postfix
}) {
  return core$1.config.nameGenerator.getName(key, '', {
    postfix
  }).join('');
}

class Processor {
  constructor({
    config
  }) {
    this.config = config;

    if (config.typescript) {
      this.tsProject = new TSProcessor(config.typescript === true ? {} : config.typescript);
    }
  }

  prepare({
    mixin
  }) {
    this.mixin = mixin;
    this.isStatic = false;
    this.variables = [];
    this.classNameNodes = new WeakSet();
    this.optimizationPaths = new Set();
  }

  isClassNameNode(node) {
    return this.classNameNodes.has(node);
  }

  classNamesToNode(classNames) {
    const properties = [];

    for (const key in classNames) {
      const classValue = classNames[key];
      const propName = t.identifier(key);
      const propValue = typeof classNames[key] === 'boolean' ? t.booleanLiteral(classValue) : t.stringLiteral(String(classValue));
      const node = t.objectProperty(propName, propValue);
      this.classNameNodes.add(node);
      properties.push(node);
    }

    return properties;
  }

  stylesToNode(key, value, {
    postfix = ''
  } = {}) {
    const {
      className
    } = taddy.$css({
      [key]: value
    }, {
      postfix
    });
    return this.classNamesToNode(className);
  }

  processPropertyValue(key, path, {
    postfix,
    properties
  }) {
    const keyPath = path.get('key');
    const valuePath = path.get('value');

    const tryLiteralValue = () => {
      if (!valuePath.isLiteral()) return false;
      const value = getLiteralValue(valuePath);
      properties.push(...this.stylesToNode(key, value, {
        postfix
      }));
      return true;
    };

    const tryObjectExpressionValue = () => {
      if (!valuePath.isObjectExpression()) return false;
      this.process(valuePath, {
        postfix: postfix + key
      });
      let currentStyles = [];

      for (const nestedProp of valuePath.node.properties) {
        if (this.isClassNameNode(nestedProp)) {
          if (currentStyles.length) {
            properties.push(t.objectProperty(keyPath.node, t.objectExpression(currentStyles)));
            currentStyles = [];
          }

          properties.push(nestedProp);
        } else {
          currentStyles.push(nestedProp);
        }
      }

      if (currentStyles.length) {
        properties.push(t.objectProperty(keyPath.node, t.objectExpression(currentStyles)));
      }

      return true;
    };

    const tryEvaluateValue = () => {
      if (!this.config.evaluate) return false;
      const {
        value
      } = evaluate(valuePath);
      if (!value) return false;
      properties.push(...this.stylesToNode(key, value, {
        postfix
      }));
      this.optimizationPaths.add(path);
      return true;
    };

    const STOP = Symbol('STOP');

    const getTSValue = tsValuePath => {
      var _this$tsProject;

      const tsType = (_this$tsProject = this.tsProject) === null || _this$tsProject === void 0 ? void 0 : _this$tsProject.getTypeAtPos(this.config.filename, this.config.code, tsValuePath.node.start || 0, tsValuePath.node.end || 0);
      if (!tsType) return;
      return parseValue(tsType.getType());
    };

    const getComposesTSValue = path => {
      const elements = [];
      elements[STATIC_KEY] = true;

      for (let elementPath of path.get('elements')) {
        if (!elementPath) return;
        const value = getTSValue(elementPath);
        if (value === STOP) return;
        elements[STATIC_KEY] = elements[STATIC_KEY] && isStaticValue(value);
        elements.push(value);
      }

      return elements;
    };

    const tryTypescriptValue = () => {
      if (!this.config.typescript) return false;
      let tsValue;

      if (key === 'composes' && valuePath.isArrayExpression()) {
        tsValue = getComposesTSValue(valuePath);
      } else {
        tsValue = getTSValue(valuePath);
      }

      if (!tsValue) return false;

      if (isStaticValue(tsValue)) {
        properties.push(...this.stylesToNode(key, tsValue, {
          postfix
        }));
        this.optimizationPaths.add(path);
        return true;
      } // TODO: improve composes handling


      if (key === 'composes') return false;
      const variants = [].concat(tsValue);
      let isCompilable = true; // pre-generate classes for variants

      variants.forEach(v => {
        if (v instanceof DynamicType) {
          isCompilable = false;
          return;
        }

        taddy.$css({
          [key]: v
        }, {
          postfix
        });
      });
      if (!isCompilable) return false;
      const propKey = getHashedName(key, {
        postfix
      });
      properties.push(t.objectProperty(t.identifier(propKey), t.callExpression(t.identifier('css.h'), [valuePath.node])));
      return true;
    };

    const tryCSSVariableValue = () => {
      if (!this.config.CSSVariableFallback) return false; // think about dynamic nested
      // in case of nested dynamic values in mixin we can't fallback them as custom properties

      if (this.mixin && postfix !== '' || key === 'composes' || key[0] === ':' || /[\W]/.test(key)) {
        return false;
      }

      const value = getHashedName(key, {
        postfix
      });
      const dynamicValue = `--${value}`;
      this.variables.push(t.objectProperty(t.stringLiteral(dynamicValue), valuePath.node));
      properties.push(...this.stylesToNode(key, `var(${dynamicValue})`, {
        postfix
      }));
      return true;
    };

    if (tryLiteralValue() || tryObjectExpressionValue() || tryEvaluateValue() || tryTypescriptValue() || tryCSSVariableValue()) {
      return;
    }

    properties.push(path.node);
  }

  processObjectProperty(path, {
    postfix = '',
    properties
  }) {
    const key = String(getObjectPropertyKey(path));

    if (key === 'className' || key === 'style' || key === null) {
      properties.push(path.node);
      return;
    }

    this.processPropertyValue(key, path, {
      postfix,
      properties
    });
  }

  processSpreadElement(path, {
    postfix = '',
    properties
  }) {
    const tryEvaluateSpreadElement = () => {
      if (!this.config.evaluate) return false;
      const {
        value
      } = evaluate(path.get('argument'));
      if (!value) return false;
      const {
        className
      } = taddy.$css(value, {
        postfix
      });
      properties.push(...this.classNamesToNode(className));
      this.optimizationPaths.add(path);
      return true;
    };

    const tryTypescriptSpreadElement = () => {
      var _this$tsProject2;

      if (!this.config.typescript) return false;
      const {
        argument
      } = path.node;
      const tsType = (_this$tsProject2 = this.tsProject) === null || _this$tsProject2 === void 0 ? void 0 : _this$tsProject2.getTypeAtPos(this.config.filename, this.config.code, argument.start || 0, argument.end || 0);
      if (!tsType) return false;
      const styles = parseValue(tsType.getType());
      if (!isStaticValue(styles)) return false;
      const {
        className
      } = taddy.$css(styles, {
        postfix
      });
      properties.push(...this.classNamesToNode(className));
      this.optimizationPaths.add(path);
      return true;
    };

    if (tryEvaluateSpreadElement() || tryTypescriptSpreadElement()) {
      return;
    }

    this.process(path.get('argument'));
    properties.push(path.node);
  }

  processLogicalExpression(path, options) {
    this.process(path.get('left'), options);
    this.process(path.get('right'), options);
  }

  processObjectExpression(path, {
    postfix = ''
  }) {
    const properties = [];

    for (const propPath of path.get('properties')) {
      if (propPath.isObjectProperty()) {
        this.processObjectProperty(propPath, {
          postfix,
          properties
        });
        continue;
      }

      if (propPath.isSpreadElement()) {
        this.processSpreadElement(propPath, {
          postfix,
          properties
        });
        continue;
      }

      properties.push(propPath.node);
    }

    path.node.properties = mergeObjectProperties(properties);
  }

  process(path, options = {}) {
    if (path.isLogicalExpression()) {
      this.processLogicalExpression(path, options);
      return;
    }

    if (path.isObjectExpression()) {
      this.processObjectExpression(path, options);
      return;
    }
  }

  run(path, {
    mixin = false
  } = {}) {
    // supports only object expressions at the moment
    if (!path.isObjectExpression()) {
      return null;
    }

    this.prepare({
      mixin
    });
    this.process(path);

    if (this.variables.length > 0) {
      path.node.properties.push(t.objectProperty(t.identifier(core$1.VARS_KEY), t.objectExpression(this.variables)));
    }

    const isStatic = !this.mixin && path.node.properties.every(x => {
      if (!t.isObjectProperty(x)) return false;

      if ('name' in x.key && x.key.name === 'composes') {
        return false;
      }

      return true;
    });

    if (isStatic) {
      optimizeStaticStyles(path);
    }

    return {
      isStatic,
      optimizationPaths: this.optimizationPaths
    };
  }

}

let LAST_INDEX = 0;
let STYLES = [];

function getStylesState() {
  const {
    rules
  } = taddy.ruleInjector.styleSheet;
  const newRules = rules.slice(LAST_INDEX);
  LAST_INDEX = rules.length;
  const added = [...newRules].map(rule => rule.cssText || '');
  STYLES.push(...added);
  return {
    added,
    total: STYLES
  };
}

function getRelativeFilepath(from, to) {
  return './' + path.relative(path.dirname(from), to);
}

function writeDevEntry({
  styles,
  jsFilepath
}) {
  const template = `
var getStyleNodeById = require('taddy').getStyleNodeById
var STYLES = '${styles}'
var state = {current: STYLES}
var isBrowser = typeof window !== 'undefined'
if (isBrowser && !getStyleNodeById('taddy').innerHTML) {
    state.current = window.__TADDY_STYLES__ = window.__TADDY_STYLES__ || STYLES
    getStyleNodeById('taddy').innerHTML = state.current
}
module.exports.STYLES = STYLES;
module.exports.state = state;
module.exports.taddy_invalidate = function (newStyles) {
    if (!newStyles) return
    state.current += newStyles
    if (isBrowser) getStyleNodeById('taddy').innerHTML += newStyles
}
`;
  fs.writeFile(jsFilepath, template, error => {
    if (error) {
      console.error('TADDY WRITEFILE ERROR', error);
    }
  });
}

function writeProdEntry({
  styles,
  jsFilepath,
  cssFilepath,
  extractCSS
}) {
  let template;

  if (extractCSS) {
    const jsToCSS = getRelativeFilepath(jsFilepath, cssFilepath);
    template = `require('${jsToCSS}');`;
  } else {
    template = `
var getStyleNodeById = require('taddy').getStyleNodeById
var STYLES = '${styles}'
if (typeof window !== 'undefined') {
    getStyleNodeById('taddy').innerHTML = STYLES
}
module.exports.STYLES = STYLES
`;
  }

  if (contentMap.has(template)) {
    return;
  }

  contentMap.set(template, true);
  fs.writeFile(jsFilepath, template, error => {
    if (error) {
      console.error('TADDY WRITEFILE ERROR', error);
    }
  });
}

function getMtime(filepath) {
  return Number(fs.statSync(filepath).mtime);
}

const filesMap = /*#__PURE__*/new Map();
const contentMap = /*#__PURE__*/new Map();

function readFileSync(filepath) {
  let data = '';

  try {
    const mtime = getMtime(filepath);

    if (filesMap.has(mtime)) {
      return filesMap.get(mtime);
    }

    data = fs.readFileSync(filepath).toString();
    filesMap.set(mtime, data);
    contentMap.set(data, mtime);
  } catch (e) {}

  return data;
}

function writeCSSFileSync(filepath, content, append) {
  const code = content + append;

  if (contentMap.has(code)) {
    return;
  }

  fs.appendFileSync(filepath, append);
}

function output({
  env,
  state,
  config: {
    // getCSSFilename = (content) => `atoms-${stringHash(content)}.css`,
    getCSSFilename = () => `atoms.css`,
    getCSSFilepath = filename => path.join(cacheDir, filename),
    // getJSFilename = (content) => `entry-${stringHash(content)}.js`,
    getJSFilename = hash => {
      return `entry` + (hash ? '.' + hash : '') + '.js';
    },
    getJSFilepath = filename => path.join(cacheDir, filename),
    relativeEntryPath = false,
    extractCSS = false
  } = {}
}) {
  const {
    added
  } = getStylesState();
  const cssFilename = getCSSFilename();
  const cssFilepath = getCSSFilepath(cssFilename);
  const stylesData = readFileSync(cssFilepath); // TODO: improve filter performance

  const diffStyles = added.filter(x => !stylesData.includes(x)).join('');
  let result = stylesData + diffStyles;
  writeCSSFileSync(cssFilepath, stylesData, diffStyles);
  const program = state.file.path;
  const {
    filename
  } = state;

  function getModulePath(jsFilepath) {
    return relativeEntryPath ? getRelativeFilepath(filename, jsFilepath) : getCachedModuleFilepath(path.basename(jsFilepath));
  }

  if (env === 'development') {
    const hash = String(stringHash(result));
    const jsFilename = getJSFilename(hash);
    const jsFilepath = getJSFilepath(jsFilename);
    writeDevEntry({
      jsFilepath,
      cssFilepath,
      extractCSS,
      styles: result
    });
    const modulePath = getModulePath(jsFilepath);
    const updaterNameNode = helperModuleImports.addNamed(program, 'taddy_invalidate', modulePath);
    program.pushContainer('body', t.callExpression(updaterNameNode, [t.stringLiteral(diffStyles), t.stringLiteral(hash)]));
    return;
  }

  const jsFilename = getJSFilename();
  const jsFilepath = getJSFilepath(jsFilename);
  writeProdEntry({
    jsFilepath,
    cssFilepath,
    extractCSS,
    styles: result
  });
  const modulePath = getModulePath(jsFilepath);
  helperModuleImports.addSideEffect(program, modulePath);
}

function isPathRemoved(path) {
  do {
    if (path.removed) return true;
  } while (path = path.parentPath);

  return false;
}

const createProcessors = config => {
  const mixinsQueue = [];
  const proceedPaths = [];
  const processor = new Processor({
    config
  });
  const handlers = {
    css: (path, {
      mixin = false
    } = {}) => {
      if (path.isMemberExpression() && 'name' in path.node.property && path.node.property.name === 'mixin') {
        mixinsQueue.push(path.parentPath);
        return;
      }

      if (!path.isCallExpression()) return;
      const argPath = path.get('arguments')[0];
      const result = processor.run(argPath, {
        mixin
      });
      if (!result) return;
      const {
        isStatic,
        optimizationPaths
      } = result;
      proceedPaths.push({
        isStatic,
        optimizationPaths
      });
      path.node.arguments = [argPath.node];
    },
    mixin: path => handlers.css(path, {
      mixin: true
    })
  };
  return {
    handlers,

    finish() {
      let isStatic = true;

      for (let x of proceedPaths) {
        isStatic = isStatic && x.isStatic;

        for (let path of x.optimizationPaths) {
          optimizeBindings(path);
        }
      }

      for (let path of mixinsQueue) {
        if (isPathRemoved(path)) {
          // the path binding should be already optimized
          continue;
        }

        handlers.css(path, {
          mixin: true
        });
      }

      return {
        isStatic
      };
    }

  };
};

function mapCompileOptions({
  filename,
  code,
  evaluate = true,
  typescript = /\.tsx?$/.test(filename),
  unstable_CSSVariableFallback = true
}) {
  return {
    filename,
    code,
    evaluate,
    typescript,
    CSSVariableFallback: unstable_CSSVariableFallback
  };
}

function macro({
  references,
  state,
  config: _config = {}
}) {
  const config = _extends({
    env: "development"
  }, _config);

  if (isTaddyEvaluation(state)) {
    return;
  }

  const program = state.file.path;
  const code = state.file.code;
  const {
    filename
  } = state;
  const {
    handlers,
    finish
  } = createProcessors(mapCompileOptions(_extends({}, config.compileOptions, {
    filename,
    code
  })));

  for (const key in references) {
    if (!handlers[key]) continue;
    references[key].forEach(path => handlers[key](path.parentPath));
  }

  let importPath;
  program.traverse({
    ImportDeclaration(p) {
      const {
        node
      } = p;

      if (!(node.source.value === PACKAGE_NAME || node.source.value === MACRO_NAME)) {
        return;
      }

      importPath = p;
      importPath.node.source = t.stringLiteral('taddy');
      p.stop();
    }

  });
  const {
    isStatic
  } = finish();

  if (isStatic && importPath) {
    importPath.node.source = t.stringLiteral('@taddy/core');
  }

  output({
    env: config.env,
    state,
    config: config.outputOptions
  });
  return {
    keepImports: true
  };
}

function getImportNames(path) {
  const localName = path.node.local.name;

  if (path.isImportSpecifier()) {
    return {
      localName,
      importedName: path.node.imported.name
    };
  }

  if (path.isImportDefaultSpecifier()) {
    return {
      localName,
      importedName: 'default'
    };
  }

  return {
    localName,
    importedName: 'namespace'
  };
}

function findReferences(paths) {
  const references = {};

  for (let path of paths) {
    const {
      localName,
      importedName
    } = getImportNames(path);
    const binding = path.scope.getBinding(localName);

    if (binding === null || binding === void 0 ? void 0 : binding.referencePaths) {
      references[importedName] = binding.referencePaths;
    }
  }

  return references;
}

function plugin(babel, options) {
  return {
    name: 'taddy',

    pre() {
      this.references = {};
    },

    visitor: {
      // TODO: support require expression
      ImportDeclaration(path, state) {
        if (isTaddyEvaluation(state)) {
          return;
        }

        if (path.node.source.value !== PACKAGE_NAME) {
          return;
        }

        Object.assign(this.references, findReferences(path.get('specifiers')));

        if (Object.keys(this.references).length === 0) {
          return;
        }

        macro({
          references: this.references,
          babel,
          state,
          config: options
        });
      }

    }
  };
}

exports.default = plugin;
exports.macro = macro;
//# sourceMappingURL=babel-plugin.cjs.development.js.map
